using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Security.Principal;
using System.Text;
using System.Threading;
using InfiniTec.Threading.Properties;

namespace InfiniTec.Threading
{
    /// <summary>
    /// This is the base class for asynchronous operations. Most of the functionality is implemented in this class.
    /// </summary>
    public partial class AsyncOperation: IDisposable
    {
        private static int _IdCounter;

        /// <summary>
        /// Called, when the async operation has completed.
        /// </summary>
        public event EventHandler<CompletedEventArgs> Completed;
        /// <summary>
        /// This event is triggered when an exception was thrown by the asynchronous operation.
        /// </summary>
        /// <remarks>If the target delegate is on a Windows form or another
        /// object that implements <see cref="System.ComponentModel.ISynchronizeInvoke"/>, the delegate is called
        /// from the thread the target object was created. In other words, window controls can be modified
        /// without a call to <see cref="M:System.Windows.Forms.Control.Invoke" />.</remarks>
        public event EventHandler<ExceptionEventArgs> Error;
        /// <summary>
        /// This event is triggered when the asynchronous operation times out.
        /// </summary>
        /// <remarks>If the target delegate is on a Windows form or another
        /// object that implements <see cref="System.ComponentModel.ISynchronizeInvoke"/>, the delegate is called
        /// from the thread the target object was created. In other words, window controls can be modified
        /// without a call to <see cref="M:System.Windows.Forms.Control.Invoke" />.</remarks>
        public event EventHandler TimedOut;

        private bool _PropagateProgressToRootOperation;
        private Exception _Exception;
        private IEnumerator<OperationAction> _OperationEnumerator;
        private bool _PropagateTimeoutToParentOperation;
        private OperationStatus _Status;
        private TimeSpan _Timeout;
        private OperationTreeNode _OperationNode;
        private int _Id = Interlocked.Increment(ref _IdCounter);

        private SynchronizationContext _SynchronizationContext = SynchronizationContext.Current ??
                                                                 new SynchronizationContext();


        /// <summary>
        /// Gets the synchronization context for the current operation.
        /// </summary>
        /// <value>The synchronization context.</value>
        protected SynchronizationContext SynchronizationContext
        {
            get { return _SynchronizationContext; }
        }

        /// <summary>
        /// True, if a timeout should be handled by the parent operation.
        /// False, otherwise.
        /// </summary>
        internal protected bool PropagateTimeoutToParentOperation
        {
            get { return _PropagateTimeoutToParentOperation; }
            set { _PropagateTimeoutToParentOperation = value; }
        }

        /// <summary>
        /// True, if progress updates should be propagated to the root operation.
        /// </summary>
        internal protected bool PropagateProgressToRootOperation
        {
            get { return _PropagateProgressToRootOperation; }
            set { _PropagateProgressToRootOperation = value; }
        }

        /// <summary>
        /// If the <see cref="Status"/> is <see cref="OperationStatus.Error"/>, this property
        /// contains the exception object that occured.
        /// </summary>
        public Exception Exception
        {
            get { return _Exception; }
        }

        internal OperationTreeNode OperationNode
        {
            get { return _OperationNode; }
        }

        /// <summary>
        /// Invokes the <see cref="Error"/> event.
        /// </summary>
        /// <param name="ex">The exception to pass to the error handler.</param>
        protected virtual void OnError(Exception ex)
        {
            if (Error != null)
                Error(this, new ExceptionEventArgs(ex));
        }

        /// <summary>
        /// Gets or sets the enumerator for the current operation.
        /// </summary>
        protected IEnumerator<OperationAction> OperationEnumerator
        {
            get { return _OperationEnumerator; }
            set { _OperationEnumerator = value; }
        }


        /// <summary>
        /// Initiates a new instance of this class with the given operation and timeout.
        /// </summary>
        /// <param name="operation">The async operation to process.</param>
        /// <param name="timeout">The timeout for this operation.</param>
        public AsyncOperation(IEnumerable<OperationAction> operation, TimeSpan timeout)
        {
            _StatusChangedSignal = new AutoResetEvent(false);
            OperationEnumerator = operation.GetEnumerator();
            _Timeout = timeout;
            _OperationNode = new OperationTreeNode(this);
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="AsyncOperation"/> class.
        /// </summary>
        /// <param name="operation">The async operation to process.</param>
        public AsyncOperation(IEnumerable<OperationAction> operation) : this(operation, TimeSpan.MaxValue) { }

        /// <summary>
        /// Signals a new state in the current operation.
        /// </summary>
        protected void SignalStatusChanged()
        {
            _StatusChangedSignal.Set();
        }
        private AutoResetEvent _StatusChangedSignal;

        /// <summary>
        /// Gets the waithandle associated with the current state of the operation.
        /// </summary>
        public WaitHandle StatusChangedSignal
        {
            get { return _StatusChangedSignal; }
        }
        

        /// <summary>
        /// Returns the current status of the operation
        /// </summary>
        public OperationStatus Status
        {
            get { return _Status; }
            protected set
            {
                _Status = value;
                SignalStatusChanged();
            }
        }

        /// <summary>
        /// Starts the async operation on a <see cref="ThreadPool"/> thread.
        /// </summary>
        public virtual void Run()
        {
            ThreadPool.QueueUserWorkItem(delegate
                {
                    Advance(this, false);
                });
        }

        /// <summary>
        /// Blocks the current thread until the async operation completes.
        /// </summary>
        public virtual void WaitForCompletion()
        {
            while (Status == OperationStatus.Running || Status == OperationStatus.NotStarted)
            {
                StatusChangedSignal.WaitOne(System.Threading.Timeout.Infinite, false);
            }

            EvaluateOperationResult();
        }

        /// <summary>
        /// Checks the result of the operation and throws an exception
        /// if the operation was not successful.
        /// </summary>
        public void EvaluateOperationResult()
        {
            switch (Status)
            {
                case OperationStatus.Error: throw new AsyncOperationFailedException(string.Format(Resources.AsyncOperationFailed, Exception.Message), Exception);
                case OperationStatus.Aborted: throw new AsyncOperationAbortedException(Resources.AsyncOperationWasAborted);
            }
        }

        /// <summary>
        /// Reports the specified progress on the current operation, if the type of the status equals
        /// the type of the StatusType of the current operation or if the StatusType of the current operation
        /// contains a property of the type T.
        /// </summary>
        /// <typeparam name="T">The type of the status to report.</typeparam>
        /// <param name="status">The status to report</param>
        /// <returns>True, if the progress update was reported successfully.</returns>
        protected internal virtual bool ReportProgress<T>(T status)
        {
            return false;
        }

        /// <summary>
        /// Sets the status of this operation to <see cref="OperationStatus.Aborting"/>. The abort
        /// query is passed to the asynchronous operation during the next ProgressChange notification.
        /// </summary>
        public virtual void Abort()
        {
            Status = OperationStatus.Aborting;
        }

        /// <summary>
        /// Specifies the timeout in milliseconds for the operation. 
        /// </summary>
        /// <remarks>
        /// The timer is reset after each step.
        /// The timeout can be modified during the operation. The new value is reflected at the next execution step.
        /// </remarks>
        public TimeSpan Timeout
        {
            get { return _Timeout; }
            set { _Timeout = value; }
        }

        private void ProcessTimeout()
        {
            Status = OperationStatus.TimedOut;
            OnTimedOut(EventArgs.Empty);
        }

        /// <summary>
        /// Executes the <see cref="TimedOut"/> event.
        /// </summary>
        /// <param name="e">The paremeters of this event.</param>
        protected virtual void OnTimedOut(EventArgs e)
        {
            if (TimedOut != null) TimedOut(this, e);
        }

        /// <summary>
        /// Gets the current operation action.
        /// </summary>
        /// <value>The current operation action.</value>
        protected OperationAction CurrentOperationAction
        {
            get { return _OperationEnumerator.Current; }
        }

        /// <summary>
        /// Processes the current operation action.
        /// </summary>
        protected virtual void ProcessCurrentOperationAction() { }


        /// <summary>
        /// Processes the given exception.
        /// </summary>
        /// <param name="ex">The exception to process.</param>
        protected void ProcessException(Exception ex)
        {
            _Exception = ex;
            Status = OperationStatus.Error;

            _SynchronizationContext.Send(delegate(object state)
                {
                    OnError((Exception) state);
                }, ex);
        }
  
        /// <summary>
        /// Processes the completion of the current operation.
        /// </summary>
        protected void ProcessOperationCompleted()
        {
            _SynchronizationContext.Send(delegate
                {
                    OnCompleted();
                }, null);
        }

       

        /// <summary>
        /// Calls the <see cref="Completed"/> event handler.
        /// </summary>
        protected void OnCompleted()
        {
            if (Completed != null)
            {
                Completed(this, new CompletedEventArgs(Status));
            }
        }

        #region IDisposable Members

        /// <summary>
        /// Releases all managed and unmanged resources held by this instance.
        /// </summary>
        public void Dispose()
        {
            Dispose(true);

            GC.SuppressFinalize(this);
        }

        /// <summary>
        /// Disposes all managed and unmanaged resources held by this instance.
        /// </summary>
        /// <param name="disposing">True, if the method was called by the <see cref="Dispose()"/>method, 
        /// false if it was called by a finalizer.</param>
        protected virtual void Dispose(bool disposing)
        {
            AutoResetEvent signal;
            IEnumerator<OperationAction> enumerator;
            OperationTreeNode node;

            if (disposing)
            {
                enumerator = Interlocked.Exchange(ref _OperationEnumerator, null);
                if (enumerator != null) enumerator.Dispose();

                node = Interlocked.Exchange(ref _OperationNode, null);
                if (node != null) node.UnwindOperation(false);

                signal = Interlocked.Exchange(ref _StatusChangedSignal, null);
                if (signal != null) signal.Close();
            }
        }

        #endregion

        internal void Start()
        {
            Debug.Assert(_Status == OperationStatus.NotStarted);

            _Status = OperationStatus.Started;
            Advance(this);
        }
    }
}
