using System;
using System.Collections.Generic;
using System.Text;
using System.Threading;

namespace InfiniTec.Threading
{
    /// <summary>
    /// Provides support methods for asynchronous operations - to ensure that only one asynchronous operation
    /// runs at a given time.
    /// </summary>
    public class AsyncOperationContext: IDisposable
    {
        private int _AsyncOperationPending;
        private AsyncOperation _PendingOperation;

        /// <summary>
        /// Checks whether an asynchronous operation is already in progress and throws an exception in that case.
        /// After the call to this method, this instance is marked as "async operation is pending".
        /// </summary>
        public void InitiatePendingAsyncOperation()
        {
            if (Interlocked.Exchange(ref _AsyncOperationPending, 1) == 1)
                throw new InvalidOperationException("Another async operation is currently in progress.");
        }

        /// <summary>
        /// Registers the specified operation as pending async operation to prevent it from being garbage collected.
        /// </summary>
        /// <param name="operation">The operation to register</param>
        public void RegisterPendingAsyncOperation(AsyncOperation operation)
        {
            _PendingOperation = operation;
        }

        /// <summary>
        /// Unregisters the pending operation and remove the "Async operation is pending" flag.
        /// </summary>
        public void UnregisterPendingAsyncOperation()
        {
            _PendingOperation = null;
            _AsyncOperationPending = 0;
        }

        /// <summary>
        /// Aborts a pending async operation.
        /// </summary>
        public void AbortPendingAsyncOperation()
        {
            if (_PendingOperation == null) throw new InvalidOperationException("No async operation is currently in progress");

            _PendingOperation.Abort();
        }

        #region IDisposable Members

        /// <summary>
        /// Performs application-defined tasks associated with freeing, releasing, or resetting unmanaged resources.
        /// </summary>
        public void Dispose()
        {
            Dispose(true);
            GC.SuppressFinalize(this);
        }

        /// <summary>
        /// Performs application-defined tasks associated with freeing, releasing, or resetting unmanaged resources.
        /// </summary>
        /// <param name="disposing">True, if called from the dispose method. False, if called from the Destructor.</param>
        protected virtual void Dispose(bool disposing)
        {
            if (!disposing) return;

            UnregisterPendingAsyncOperation();
        }

        #endregion

    }
}
