using System;
using System.Threading;

namespace InfiniTec.Threading
{
    /// <summary>
    /// Implements base functionality for classes which implement asynchronous operations.
    /// </summary>
    public abstract class AsyncOperationObject
    {
        private int _AsyncOperationPending;
        private AsyncOperation _PendingOperation;

        /// <summary>
        /// Checks whether an asynchronous operation is already in progress and throws an exception in that case.
        /// After the call to this method, this instance is marked as "async operation is pending".
        /// </summary>
        protected void InitiatePendingAsyncOperation()
        {
            if (Interlocked.Exchange(ref _AsyncOperationPending, 1) == 1)
                throw new InvalidOperationException("Another async operation is currently in progress.");
        }

        /// <summary>
        /// Registers the specified operation as pending async operation to prevent it from being garbage collected.
        /// </summary>
        /// <param name="operation">The operation to register</param>
        protected void RegisterPendingAsyncOperation(AsyncOperation operation)
        {
            _PendingOperation = operation;
        }

        /// <summary>
        /// Unregisters the pending operation and remove the "Async operation is pending" flag.
        /// </summary>
        protected AsyncOperation UnregisterPendingAsyncOperation()
        {
            AsyncOperation result;

            result = _PendingOperation;
            _PendingOperation = null;

            Interlocked.Exchange(ref _AsyncOperationPending, 0);
            return result;
        }

        /// <summary>
        /// Aborts a pending async operation.
        /// </summary>
        public void AbortPendingAsyncOperation()
        {
            if (_PendingOperation == null) throw new InvalidOperationException("No async operation is currently in progress");

            _PendingOperation.Abort();
        }
    }
}
