using System;
using System.Collections.Generic;
using System.Text;
using System.Threading;

namespace InfiniTec.Threading
{
    partial class AsyncOperation
    {
        /// <summary>
        /// Processes the next step of the active operation.
        /// </summary>
        /// <param name="operation">The <see cref="AsyncOperation"/> to be processed.</param>
        protected internal static void Advance(AsyncOperation operation)
        {
            Advance(operation, false);
        }

        /// <summary>
        /// Processes the next step of the active operation.
        /// </summary>
        /// <param name="state">Not used</param>
        /// <param name="timedOut">True, if the operation timed out. False, otherwise.</param>
        private static void Advance(object state, bool timedOut)
        {
            AsyncOperation operation;

            operation = state as AsyncOperation;

            if (operation == null) throw new ArgumentNullException("state");
            do
            {
                do
                {
                    try
                    {
                        if (timedOut)
                        {
                            // The operation timed out.

                            operation = PropagateTimeout(operation);

                            if (operation != null) continue;
                            else break;
                        }

                        if (operation.Status == OperationStatus.Aborting)
                        {
                            AbortOperation(operation);
                            break;
                        }

                        operation._Status = OperationStatus.Running;
                        // Perform the next step of the operation
                        if (operation.OperationEnumerator.MoveNext())
                        {
                            bool exitMethod;

                            exitMethod = ProcessOperationAction(operation);

                            if (exitMethod) return;
                        }
                        else
                        {
                            operation.ProcessOperationCompleted();
                            operation.Status = OperationStatus.Completed;

                            break;
                        }
                    }
                    catch (Exception ex)
                    {
                        // Bubble exception up to a point where it can be handled.

                        operation = PropagateException(ex, operation);

                        if (operation != null) continue;
                        else return;
                    }
                } while (true);

                operation = UnwindOperation(operation);

                if (operation != null && !operation.OperationNode.IsLeaf) return;

            } while (operation != null);
        }

        private static void AbortOperation(AsyncOperation operation)
        {
            operation.Dispose();

            operation._Status = OperationStatus.Aborted;

            operation.ProcessOperationCompleted();
        }

        private static bool ProcessOperationAction(AsyncOperation operation)
        {
            OperationAction action;

            operation.ProcessCurrentOperationAction();

            action = operation.CurrentOperationAction;

            if (action is NestedOperation)
            {
                ProcessNestedOperationAction((NestedOperation)action, operation, true);

                return true;
            }

            else if (action is ParallelOperations)
            {
                ProcessParallelOperations((ParallelOperations)action, operation);
                return true;
            }
            else if (action is ProgressUpdate)
            {
                return ProcessProgressUpdateAction(operation);
            }

            return false;
 
        }

        private static bool ProcessProgressUpdateAction(AsyncOperation operation)
        {
            WaitHandle asyncWaitHandle;
            int timeout;
            ProgressUpdate update;

            update = (ProgressUpdate) operation.CurrentOperationAction;

            asyncWaitHandle = update.AsyncWaitHandle;
            
            if (asyncWaitHandle != null)
            {
                timeout = (operation.Timeout == TimeSpan.MaxValue) ? -1 : (int) operation.Timeout.TotalMilliseconds;
                    
                ThreadPool.RegisterWaitForSingleObject(
                    asyncWaitHandle, Advance, operation, timeout, true);
                return true;
            }
        
            return false;
        }


        private static AsyncOperation PropagateException(Exception ex, AsyncOperation operation)
        {
            bool exceptionHandled;

            do
            {
                exceptionHandled = operation.OperationNode.HandlesException(ex);

                operation.ProcessException(ex);

                operation = UnwindOperation(operation);
            } while (operation != null && !exceptionHandled);

            return operation;
        }

        private static AsyncOperation PropagateTimeout(AsyncOperation operation)
        {
            if (operation != null)
            {
                while (operation != null &&
                    !operation.PropagateTimeoutToParentOperation)
                {
                    operation = UnwindOperation(operation);
                }
            }
            if (operation != null)
            {
                operation.ProcessTimeout();
            }

            return operation;
        }

        private static AsyncOperation UnwindOperation(AsyncOperation operation)
        {
            OperationTreeNode node;

            node = Interlocked.Exchange(ref operation._OperationNode, null);
            
            if (node == null) return null;

            node.UnwindOperation(true);

            if (node.Parent != null)
            {
                return node.Parent.Operation;
            }
            else 
            {
                return null;
            }
        }

        /// <summary>
        /// Pushes the <see cref="NestedOperation"/> onto the operation stack and executes it.
        /// </summary>
        /// <param name="action">The operation to execute.</param>
        /// <param name="parentOperaton">The parent operation</param>
        /// <param name="startOperation">If set to true, the nested action is executed immediately.</param>
        static void ProcessNestedOperationAction(NestedOperation action, AsyncOperation parentOperaton, bool startOperation)
        {
            AsyncOperation nestedOperation;
            OperationTreeNode currentNode;

            nestedOperation = action.Operation;

            currentNode = parentOperaton.OperationNode;

            currentNode.AddSubOperation(action.Operation, action.ExceptionsToCatch);

            nestedOperation.PropagateProgressToRootOperation = action.PropagateProgressToRootOperation;
            nestedOperation.PropagateTimeoutToParentOperation = action.PropagateTimeout;

            if (startOperation) parentOperaton.OperationNode.StartNestedOperations();
        }

        private static void ProcessParallelOperations(ParallelOperations operations, AsyncOperation parentOperation)
        {
            parentOperation.OperationNode.MaxConcurrentOperations = operations.MaxConcurrentOperations;

            foreach (NestedOperation operation in operations.Operations)
            {
                ProcessNestedOperationAction(operation, parentOperation, false);
            }

            parentOperation.OperationNode.StartNestedOperations();
        }
    }
}
