using System;
using System.Collections.Generic;
using System.Text;
using System.Threading;
using System.Runtime.InteropServices;
using System.Reflection;

namespace InfiniTec.Threading
{
    /// <summary>
    /// This class aids in development of asynchronous operations.
    /// </summary>
    /// <remarks>
    /// The Operation itself is implemented as an iterator. Every step in the operation which can be performed
    /// asynchronously (using the BeginXXX methods) will be complemented with a yield return of 
    /// <see cref="ProgressUpdate"/>, passing along the waithandle retrieved by the BeginXXX method.
    /// When the <see cref="WaitHandle"/> is signaled, the operation is resumed at that position. This makes it
    /// possible to call the EndXXX method in the same method as the BeginXXX method.
    /// </remarks>
    /// <example>
    /// The following code show the usage for a single step in the operation:
    /// <code>
    ///  WebRequest request = WebRequest.Create(url);
    ///  ar = request.BeginGetResponse(null, null);
    /// 
    ///  action = new ProgressChangedStateBag&lt;Status&gt;(Status.GetResponse, ar.AsyncWaitHandle);
    ///  yield return action;
    ///  if (action.Cancel == true) yield break;
    /// 
    ///  WebResponse response = request.EndGetResponse(ar);
    /// 
    /// </code>
    /// <para>The idea for this class came originally from Michael Entin.</para>
    /// </example>
    /// <typeparam name="StatusType">The type used for status reports.</typeparam>
    public class AsyncOperation<StatusType>: AsyncOperation
    {
        /// <summary>
        /// This event is triggered when the progress of the asynchronous operation changes.
        /// </summary>
        public event ProgressChangedEventHandler<StatusType> ProgressChanged;

        private Dictionary<Type, PropertyInfo> _NestedStateMembers;

        private StatusType _StatusInformation;

        /// <summary>
        /// Returns the status object from the last ProgressChanged notification.
        /// </summary>
        public StatusType StatusInformation
        {
            get { return _StatusInformation; }
            private set { _StatusInformation = value; }
        }

        private int _ProgressPercentage;



        /// <summary>
        /// Returns the current progress of the operation in percent.
        /// </summary>
        public int ProgressPercentage
        {
            get { return _ProgressPercentage; }
            private set { _ProgressPercentage = value; }
        }

        /// <summary>
        /// Invokes the <see cref="ProgressChanged"/> event.
        /// </summary>
        /// <param name="e">The event information that is passed to the event handler</param>
        protected virtual void OnProgressChanged(ProgressChangedEventArgs<StatusType> e)
        {
            if (ProgressChanged != null)
                ProgressChanged(this, e);
        }

        /// <summary>
        /// Initializes a new instance of this class for the specified operation.
        /// </summary>
        /// <param name="operation">The asynchronous operation to run.</param>
        /// <param name="timeout">The timeout for each step of the operation, in milliseconds</param>
        public AsyncOperation(IEnumerable<OperationAction> operation, TimeSpan timeout): 
            base(operation, timeout)
        {
            _NestedStateMembers = new Dictionary<Type, PropertyInfo>();
        }

          /// <summary>
        /// Initializes a new instance of this class for the specified operation.
        /// </summary>
        /// <param name="operation">The asynchronous operation to run.</param>
        public AsyncOperation(IEnumerable<OperationAction> operation)
            : this(operation, TimeSpan.MaxValue)
        {
        }

        /// <summary>
        /// Processes the current operation action.
        /// </summary>
        protected override void ProcessCurrentOperationAction()
        {
            // Report progress
            if (CurrentOperationAction is ProgressUpdate<StatusType>)
            {
                ProcessProgressChangedBag((ProgressUpdate<StatusType>)CurrentOperationAction);
            }
            else if (CurrentOperationAction is NestedOperation<StatusType>)
            {
                ProcessProgressChangedBag(((NestedOperation<StatusType>)CurrentOperationAction).Status);
                base.ProcessCurrentOperationAction();
            }
            else if (CurrentOperationAction is OperationResult<StatusType>)
            {
                ProcessResultBag((OperationResult<StatusType>)CurrentOperationAction);
            }
            else
            {
                base.ProcessCurrentOperationAction();
            }
        }

        private void ProcessResultBag(OperationResult<StatusType> result)
        {
            ProgressUpdate<StatusType> update;

            update = new ProgressUpdate<StatusType>(result.Result);
            update.ProgressPercentage = 100;

            ProcessProgressChangedBag(update);

            Status = OperationStatus.Completed;
        }

        /// <summary>
        /// Reports the specified progress on the current operation, if the type of the status equals
        /// the type of the StatusType of the current operation or if the StatusType of the current operation
        /// contains a property of the type T.
        /// </summary>
        /// <typeparam name="T">The type of the status to report.</typeparam>
        /// <param name="status">The status to report</param>
        /// <returns>True, if the progress update was reported successfully.</returns>
        protected internal override bool ReportProgress<T>(T status)
        {
            PropertyInfo info;
            ProgressChangedEventArgs<StatusType> progress;

            if (typeof(T) == typeof(StatusType))
            {
                _StatusInformation = (StatusType) ((object) status);
            }
            else
            {

                if (StatusInformation == null) return false;

                info = GetNestedStateProperty<T>();

                if (info == null) return false;

                info.SetValue(StatusInformation, status, null);
            }
            progress = new ProgressChangedEventArgs<StatusType>(StatusInformation, ProgressPercentage);

            SynchronizationContext.Send(delegate(object state)
                {
                    OnProgressChanged((ProgressChangedEventArgs<StatusType>) state);
                }, progress); 
            return progress.Cancel;
        }
        /// <summary>
        /// Reports the specified progress on the current operation, if the type of the status equals
        /// the type of the StatusType of the current operation or if the StatusType of the current operation
        /// contains a property of the type T.
        /// </summary>
        /// <typeparam name="T">The type of the status to report.</typeparam>
        /// <returns>True, if the progress update was reported successfully.</returns>
        private PropertyInfo GetNestedStateProperty<T>()
        {
            Type statusType;
            PropertyInfo info;

            if (_NestedStateMembers.ContainsKey(typeof(T))) return _NestedStateMembers[typeof(T)];

            statusType = typeof(StatusType);

            info = Array.Find(statusType.GetProperties(), delegate(PropertyInfo pi)
            {
                if (!NestedStateAttribute.IsDefined(pi)) return false;
                return (pi.PropertyType == typeof(T));
            });

            _NestedStateMembers[typeof(T)] = info;

            return info;
        }

        private void ProcessProgressChangedBag(ProgressUpdate<StatusType> statusBag)
        {
            ProgressChangedEventArgs<StatusType> e;

            StatusInformation = statusBag.Status;
            ProgressPercentage = statusBag.ProgressPercentage;
            
            if (Status != OperationStatus.Aborting) Status = OperationStatus.Running;

            e = new ProgressChangedEventArgs<StatusType>(
                statusBag.Status, statusBag.ProgressPercentage);

            SynchronizationContext.Send(delegate(object state)
                {
                    OnProgressChanged((ProgressChangedEventArgs<StatusType>) state);
                }, e); 

            if (PropagateProgressToRootOperation) 
                e.Cancel = OperationNode.RootOperationNode.Operation.ReportProgress(statusBag.Status);

            if (e.Cancel || (Status == OperationStatus.Aborting))
            {
                Status = OperationStatus.Aborting;
            }
        }
    }
}
