using System;
using System.Collections.Generic;
using System.Text;
using System.Threading;

namespace InfiniTec.Threading
{
    /// <summary>
    /// This class aids in development of asynchronous operations.
    /// </summary>
    /// <remarks>
    /// The Operation itself is implemented as an iterator. Every step in the operation which can be performed
    /// asynchronously (using the BeginXXX methods) will be complemented with a yield return of 
    /// <see cref="ProgressUpdate"/>, passing along the waithandle retrieved by the BeginXXX method.
    /// When the <see cref="WaitHandle"/> is signaled, the operation is resumed at that position. This makes it
    /// possible to call the EndXXX method in the same method as the BeginXXX method.
    /// </remarks>
    /// <example>
    /// The following code show the usage for a single step in the operation:
    /// <code>
    ///  WebRequest request = WebRequest.Create(url);
    ///  ar = request.BeginGetResponse(null, null);
    /// 
    ///  action = new ProgressUpdate&lt;Status&gt;(Status.GetResponse, ar.AsyncWaitHandle);
    ///  yield return action;
    ///  if (action.Cancel == true) yield break;
    /// 
    ///  WebResponse response = request.EndGetResponse(ar);
    /// 
    /// </code>
    /// <para>The idea for this class came originally from Michael Entin.</para>
    /// </example>
    /// <typeparam name="StatusType">The type used for status reports.</typeparam>
    /// <typeparam name="ResultType">The type returned by the operation.</typeparam>
    public class AsyncOperation<StatusType, ResultType>: AsyncOperation<StatusType>
    {
        private ResultType _Result;

        /// <summary>
        /// Returns the result information from the last Completed notification.
        /// </summary>
        public ResultType Result
        {
            get { return _Result; }
            private set { _Result = value; }
        }

        /// <summary>
        /// Initializes a new instance of this class for the specified operation, with a timeout of 
        /// <see cref="System.Threading.Timeout.Infinite"/>.
        /// </summary>
        /// <param name="operation">The asynchronous operation to run.</param>
        public AsyncOperation(IEnumerable<OperationAction> operation): this(operation, TimeSpan.MaxValue)
        {
        }

        /// <summary>
        /// Initializes a new instance of this class for the specified operation.
        /// </summary>
        /// <param name="operation">The asynchronous operation to run.</param>
        /// <param name="timeout">The timeout for each step of the operation, in milliseconds</param>
        public AsyncOperation(IEnumerable<OperationAction> operation, TimeSpan timeout): 
            base(operation,  timeout)
        {
        }
        private void ProcessResultBag(OperationResult<ResultType> statusBag)
        {
            Result = statusBag.Result;
        }

        /// <summary>
        /// Processes the current operation action.
        /// </summary>
        protected override void ProcessCurrentOperationAction()
        {
            if (CurrentOperationAction is OperationResult<ResultType>)
            {
                ProcessResultBag((OperationResult<ResultType>)CurrentOperationAction);
            }
            else
            {
                base.ProcessCurrentOperationAction();
            }
        }

    }
}
