using System;
using System.Collections.Generic;
using System.Text;
using System.IO;

namespace InfiniTec.Threading
{
    /// <summary>
    /// This static class implements often used async operations. 
    /// </summary>
    public static class AsyncOperations
    {
        /// <summary>
        /// Returns an async operation that copies the contents of an input stream to an output stream.
        /// </summary>
        /// <param name="input">The stream to copy the data from.</param>
        /// <param name="output">The stream to copy the data to.</param>
        /// <param name="bytesToCopy">The number of bytes to copy.</param>
        /// <param name="bufferSize">The buffersize for each copy operation.</param>
        /// <returns></returns>
        public static AsyncOperation<CopyOperation> Copy(Stream input, Stream output, long bytesToCopy, int bufferSize)
        {
            return new AsyncOperation<CopyOperation>(CopyInternal(input, output, bytesToCopy, bufferSize));
        }
        /// <summary>
        /// Returns an async operation that copies the contents of an input stream to an output stream.
        /// </summary>
        /// <param name="input">The stream to copy the data from.</param>
        /// <param name="output">The stream to copy the data to.</param>
        /// <param name="bytesToCopy">The number of bytes to copy.</param>
        public static AsyncOperation<CopyOperation> Copy(Stream input, Stream output, long bytesToCopy)
        {
            return new AsyncOperation<CopyOperation>(CopyInternal(input, output, bytesToCopy, 4096));
        }

        /// <summary>
        /// Returns an async operation that copies the contents of an input stream to an output stream.
        /// </summary>
        /// <param name="input">The stream to copy the data from.</param>
        /// <param name="output">The stream to copy the data to.</param>
        public static AsyncOperation<CopyOperation> Copy(Stream input, Stream output)
        {
            return new AsyncOperation<CopyOperation>(CopyInternal(input, output, long.MaxValue, 4096));
        }
    
        private static IEnumerable<OperationAction> CopyInternal(Stream input, Stream output, long bytesToCopy, int bufferSize)
        {
            ProgressUpdate<CopyOperation> progress;
            byte[] buffer; 
            int bytesCopied;
            int count;
            IAsyncResult ar;

            bytesCopied = 0;

            progress = new ProgressUpdate<CopyOperation>();

            buffer = new byte[Math.Min(bufferSize, bytesToCopy)];

            do {
                ar = input.BeginRead(buffer, 0, buffer.Length, null, null);
                
                yield return progress.Update(CopyOperation.Reading, (int) ((double) bytesCopied*100 / bytesToCopy*2),
                    ar.AsyncWaitHandle);

                count = input.EndRead(ar);

                ar = output.BeginWrite(buffer, 0, count, null, null);

                yield return progress.Update(CopyOperation.Writing, 
                    (int) ((double) (bytesCopied+count)*100 / bytesToCopy), ar.AsyncWaitHandle);

                output.EndWrite(ar);
                
                bytesCopied += count;
            } while (count > 0 && bytesCopied < bytesToCopy);

        }
    }
}
