using System;
using System.Collections.Generic;
using System.Text;
using System.ComponentModel;

namespace InfiniTec.Threading
{
    /// <summary>
    /// This class is used to report the progress of an asynchronous operation.
    /// </summary>
    /// <typeparam name="StatusType">The type of the <see cref="Status"/> object.</typeparam>
    public class ProgressChangedEventArgs<StatusType> : CancelEventArgs
    {
        private StatusType _Status;

        /// <summary>
        /// The object containing the state information of the current operation.
        /// </summary>
        public StatusType Status
        {
            get { return _Status; }
            set { _Status = value; }
        }

        private int _ProgressPercentage;

        /// <summary>
        /// The progress in percent.
        /// </summary>
        public int ProgressPercentage
        {
            get { return _ProgressPercentage; }
            set { _ProgressPercentage = value; }
        }

        /// <summary>
        /// Initializes a new instance of this class.
        /// </summary>
        /// <param name="state">The current status of the operation.</param>
        /// <param name="progressPercentage">The amoun of work completed, in percent.</param>
        public ProgressChangedEventArgs(StatusType state, int progressPercentage)
        {
            _Status = state;
            _ProgressPercentage = progressPercentage;
        }
    }

    /// <summary>
    /// This class is used to indicate an error during the operation.
    /// </summary>
    public class ExceptionEventArgs : EventArgs
    {
        private Exception _Exception;

        /// <summary>
        /// The exception which occured during the operation.
        /// </summary>
        public Exception Exception
        {
            get { return _Exception; }
            set { _Exception = value; }
        }

        /// <summary>
        /// Initializes a new instance of this class using the spcecified exception.
        /// </summary>
        /// <param name="ex">The exception to report.</param>
        public ExceptionEventArgs(Exception ex)
        {
            _Exception = ex;
        }
    }

    /// <summary>
    /// This class contains information about a completed async operation.
    /// </summary>
    public class CompletedEventArgs: EventArgs
    {
        private OperationStatus _Status;

        /// <summary>
        /// Returns the status of the operation
        /// </summary>
        public OperationStatus Status
        {
            get { return _Status; }
        }

        /// <summary>
        /// Initiates a new instance of this class with the specified <see cref="OperationStatus"/>
        /// </summary>
        /// <param name="status">A valid value of the <see cref="OperationStatus"/> enumeration.</param>
        public CompletedEventArgs(OperationStatus status)
        {
            _Status = status;
        }
    }
    /// <summary>
    /// Returns the result of the operation.
    /// </summary>
    /// <typeparam name="ResultType">The type of the result object</typeparam>
    public class CompletedEventArgs<ResultType> : CompletedEventArgs
    {
        private ResultType _Result;

        /// <summary>
        /// Returns the result of the asynchronous operation.
        /// </summary>
        public ResultType Result
        {
            get { return _Result; }
        }

        /// <summary>
        /// Initializes a new instance of this class using the specified result object.
        /// </summary>
        /// <param name="result">The result of the operation</param>
        /// <param name="status">The current status of the operation</param>
        public CompletedEventArgs(ResultType result, OperationStatus status): base(status)
        {
            _Result = result;
        }
    }
}
