using System;

namespace InfiniTec.Threading
{
    /// <summary>
    /// Implements an async operation step, which executes excactly one step of the nested operation and then returns.
    /// </summary>
    public abstract class NestedOperationStep: OperationAction
    {
        private AsyncOperation _Operation;
        private Type[] _ExceptionsToCatch;

        /// <summary>
        /// Returns the types of <see cref="Exception"/> handled by the calling operation.
        /// </summary>
        public Type[] ExceptionsToCatch
        {
            get { return _ExceptionsToCatch; }
            private set
            {
                if (value == null) return;

                Array.ForEach(value, delegate(Type t)
                {
                    if (t != typeof(Exception) && !(t.IsSubclassOf(typeof(Exception))))
                        throw new ArgumentException("Only exceptions are permitted!");
                });

                _ExceptionsToCatch = value;

            }
        }

        /// <summary>
        /// Gets or sets the <see cref="AsyncOperation"/> that is to be executed.
        /// </summary>
        public AsyncOperation Operation
        {
            get { return _Operation; }
            set { _Operation = value; }
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="NestedOperationStep"/> class.
        /// </summary>
        /// <param name="operation">The operation.</param>
        /// <param name="exceptionsToCatch">The exceptions to catch.</param>
        public NestedOperationStep(AsyncOperation operation, params Type[] exceptionsToCatch)
        {
            _Operation = operation;
            ExceptionsToCatch = exceptionsToCatch;
        }

    }
}
