using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Text;
using System.Threading;

namespace InfiniTec.Threading
{
    internal class OperationTreeNode
    {
        private OperationTreeNode _Parent;
        private AsyncOperation _Operation;
        private List<OperationTreeNode> _NestedOperations;
        private object _NestedOperationsLock = new object();
        private Type[] _HandledExceptions;
        private int _MaxConcurrentOperations = int.MaxValue;
        private Queue<OperationTreeNode> _PendingOperations = new Queue<OperationTreeNode>();

        public OperationTreeNode(AsyncOperation operation)
        {
            _Operation = operation;
            _NestedOperations = new List<OperationTreeNode>();
        }

        public bool HandlesException(Exception ex)
        {
            if (_HandledExceptions == null) return false;

            return Array.Find(_HandledExceptions, delegate(Type t)
                {
                    return
                        t.IsInstanceOfType(ex);
                }) != null;
        }

        public OperationTreeNode(AsyncOperation operation, OperationTreeNode parent): this(operation)
        {
            _Operation.OperationNode._Parent = parent;
        }

        public OperationTreeNode RootOperationNode
        {
            get
            {
                OperationTreeNode node = this;

                while (node.Parent != null)
                {
                    node = node.Parent;
                }
                return node;
            }
        }
        public OperationTreeNode(AsyncOperation operation, OperationTreeNode parent, Type[] handledExceptions):
            this(operation, parent)
        {
            _HandledExceptions = handledExceptions;
        }

        public bool IsLeaf
        {
            get { return _NestedOperations.Count == 0; }
        }

        public AsyncOperation Operation
        {
            get { return _Operation; }
        }

        public OperationTreeNode Parent
        {
            get { return _Parent; }
            set { _Parent = value;}
        }

        public int MaxConcurrentOperations
        {
            get { return _MaxConcurrentOperations; }
            set { _MaxConcurrentOperations = value; }
        }

        public void AddSubOperation(AsyncOperation operation, params Type[] handledExceptions)
        {
            lock (_NestedOperationsLock)
            {
                _PendingOperations.Enqueue(UpdateOperationTreeNode(operation, handledExceptions));
            }
        }

        private OperationTreeNode UpdateOperationTreeNode(AsyncOperation operation, Type[] handledExceptions)
        {
            OperationTreeNode node;

            node = operation.OperationNode;

            Debug.Assert(node.Parent == null, "Node.parent is not null");
            Debug.Assert(node._NestedOperations.Count == 0, "child operation has already child operations");

            node._Parent = this;
            node._HandledExceptions = handledExceptions;

            return node;
        }

        public void UnwindOperation(bool disposeBaseOperation)
        {
            OperationTreeNode[] nestedOperations;

            if (!IsLeaf)
            {
                
                lock (_NestedOperationsLock)
                {
                    nestedOperations = _NestedOperations.ToArray();
                    _NestedOperations.Clear();
                }

                foreach (OperationTreeNode node in nestedOperations)
                {
                    node.UnwindOperation(true);
                }

            }

            if (disposeBaseOperation)
            {
                _Operation.Dispose();

                if (_Parent == null) return;

                _Parent.RemoveSubOperation(this);
            }
        }

        public void RemoveSubOperation(OperationTreeNode operation)
        {
            lock (_NestedOperationsLock)
            {
                _NestedOperations.Remove(operation);

                if (_PendingOperations.Count > 0)
                {
                    StartNestedOperations();
                }
            }
        }

        public void StartNestedOperations()
        {
            int max;
            Queue<OperationTreeNode> operations;
            OperationTreeNode operation;

            lock (_NestedOperationsLock)
            {
                operations = new Queue<OperationTreeNode>();
                max = Math.Min(_PendingOperations.Count, MaxConcurrentOperations);

                for (int i=0; i<max; i++)
                {
                    operation = _PendingOperations.Dequeue();
                    _NestedOperations.Add(operation);
                    operations.Enqueue(operation);
                }
            }

            while (operations.Count > 0)
            {
                operation = operations.Dequeue();

                ThreadPool.QueueUserWorkItem(delegate(object state)
                    {
                        OperationTreeNode entry;

                        entry = (OperationTreeNode) state;
                        entry.Operation.Start();

                    }, operation);
            }
        }
    }
}
