using System;
using System.Collections.Generic;
using System.Text;
using System.Threading;

namespace InfiniTec.Threading
{
    /// <summary>
    /// This action is used to report progress information during an asynchronous operation.
    /// </summary>
    /// <typeparam name="StatusType">The type of the <see cref="Status"/> property</typeparam>
    public class ProgressUpdate<StatusType>: ProgressUpdate
    {

        private StatusType _Status;

        /// <summary>
        /// Gets or sets the status object of this action
        /// </summary>
        public StatusType Status
        {
            get { return _Status; }
            set { _Status = value; }
        }

        /// <summary>
        /// Updates this instance with the specified information
        /// </summary>
        /// <param name="status">The new status of the operation</param>
        /// <param name="progressPercentage">The amount of work completed</param>
        /// <param name="waitHandle">The next <see cref="WaitHandle"/> for the operation.</param>
        public ProgressUpdate<StatusType> Update(StatusType status, int progressPercentage, WaitHandle waitHandle)
        {
            Status = status;
            ProgressPercentage = progressPercentage;
            AsyncWaitHandle = waitHandle;

            return this;
        }

        /// <summary>
        /// Updates this instance with the specified information
        /// </summary>
        /// <param name="status">The new status of the operation</param>
        /// <param name="progressPercentage">The amount of work completed</param>
        public ProgressUpdate<StatusType> Update(StatusType status, int progressPercentage)
        {
            Status = status;
            ProgressPercentage = progressPercentage;
            AsyncWaitHandle = null;

            return this;
        }

        /// <summary>
        /// Initializes a new instance of this class with the given status and waithandle, and sets the
        /// progress percentage to zero.
        /// </summary>
        /// <param name="status">The status of the current operation</param>
        /// <param name="asyncWaitHandle">The waithandle of the current state of the operation.</param>
        /// <param name="progressPercentage">The current progress in percent</param>
        public ProgressUpdate(StatusType status, WaitHandle asyncWaitHandle, int progressPercentage)
            : base(progressPercentage, asyncWaitHandle)
        {
            _Status = status;
        }

        /// <summary>
        /// Initializes a new instance of this class with the specified status, progress and waithandle.
        /// </summary>
        /// <param name="status">The status of the current operation</param>
        /// <param name="asyncWaitHandle">The waithandle of the current state of the operation.</param>
        public ProgressUpdate(StatusType status, WaitHandle asyncWaitHandle)
            :
            this(status, asyncWaitHandle, 0) { }

        /// <summary>
        /// Initializes a new instance of this class with the specified status, progress and waithandle.
        /// </summary>
        /// <param name="status">The status of the current operation</param>
        public ProgressUpdate(StatusType status)
            :
            this(status, null, 0) { }

        /// <summary>
        /// Initializes a new instance of this class with the status set to the default value for its type.
        /// The <see cref="T:System.Threading.Waithandle" /> property is set to <see langref="null" />
        /// </summary>
        public ProgressUpdate()
            : base()
        {
            _Status = default(StatusType);
        }
    }
}
